#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import collections

from osc_lib.cli import format_columns
from osc_lib.tests import utils


class TestDictColumn(utils.TestCase):
    def test_dict_column(self):
        data = {
            'key1': 'value1',
            'key2': 'value2',
        }
        col = format_columns.DictColumn(data)
        self.assertEqual(data, col.machine_readable())
        self.assertEqual("key1='value1', key2='value2'", col.human_readable())

    def test_complex_object(self):
        """Non-dict objects should be converted to a dict."""
        data = collections.OrderedDict(
            [('key1', 'value1'), ('key2', 'value2')]
        )
        col = format_columns.DictColumn(data)
        # we explicitly check type rather than use isinstance since an
        # OrderedDict is a subclass of dict and would inadvertently pass
        self.assertEqual(type(col.machine_readable()), dict)


class TestDictListColumn(utils.TestCase):
    def test_dict_list_column(self):
        data = {
            'public': ['2001:db8::8', '172.24.4.6'],
            'private': ['2000:db7::7', '192.24.4.6'],
        }
        col = format_columns.DictListColumn(data)
        self.assertEqual(data, col.machine_readable())
        self.assertEqual(
            'private=192.24.4.6, 2000:db7::7; public=172.24.4.6, 2001:db8::8',
            col.human_readable(),
        )

    def test_complex_object(self):
        """Non-dict-of-list objects should be converted to a dict-of-lists."""
        data = collections.OrderedDict(
            [('key1', ['value1']), ('key2', ['value2'])],
        )
        col = format_columns.DictListColumn(data)
        # we explicitly check type rather than use isinstance since an
        # OrderedDict is a subclass of dict and would inadvertently pass
        self.assertEqual(type(col.machine_readable()), dict)


class TestListColumn(utils.TestCase):
    def test_list_column(self):
        data = [
            'key1',
            'key2',
        ]
        col = format_columns.ListColumn(data)
        self.assertEqual(data, col.machine_readable())
        self.assertEqual("key1, key2", col.human_readable())

    def test_complex_object(self):
        """Non-list objects should be converted to a list."""
        data = {'key1', 'key2'}
        col = format_columns.ListColumn(data)
        # we explicitly check type rather than use isinstance since an
        # OrderedDict is a subclass of dict and would inadvertently pass
        self.assertEqual(type(col.machine_readable()), list)


class TestListDictColumn(utils.TestCase):
    def test_list_dict_column(self):
        data = [
            {'key1': 'value1'},
            {'key2': 'value2'},
        ]
        col = format_columns.ListDictColumn(data)
        self.assertEqual(data, col.machine_readable())
        self.assertEqual("key1='value1'\nkey2='value2'", col.human_readable())

    def test_complex_object(self):
        """Non-list-of-dict objects should be converted to a list-of-dicts."""
        # not actually a list (which is the point)
        data = (
            collections.OrderedDict([('key1', 'value1'), ('key2', 'value2')]),
        )
        col = format_columns.ListDictColumn(data)
        # we explicitly check type rather than use isinstance since an
        # OrderedDict is a subclass of dict and would inadvertently pass
        self.assertEqual(type(col.machine_readable()), list)  # noqa: H212
        for x in col.machine_readable():
            self.assertEqual(type(x), dict)  # noqa: H212


class TestSizeColumn(utils.TestCase):
    def test_size_column(self):
        content = 1576395005
        col = format_columns.SizeColumn(content)
        self.assertEqual(content, col.machine_readable())
        self.assertEqual('1.6G', col.human_readable())



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import argparse

from osc_lib.cli import parseractions
from osc_lib.tests import utils


class TestKeyValueAction(utils.TestCase):
    def setUp(self):
        super(TestKeyValueAction, self).setUp()

        self.parser = argparse.ArgumentParser()

        # Set up our typical usage
        self.parser.add_argument(
            '--property',
            metavar='<key=value>',
            action=parseractions.KeyValueAction,
            default={'green': '20%', 'format': '#rgb'},
            help='Property to store for this volume '
            '(repeat option to set multiple properties)',
        )

    def test_good_values(self):
        results = self.parser.parse_args(
            [
                '--property',
                'red=',
                '--property',
                'green=100%',
                '--property',
                'blue=50%',
            ]
        )

        actual = getattr(results, 'property', {})
        # All should pass through unmolested
        expect = {'red': '', 'green': '100%', 'blue': '50%', 'format': '#rgb'}
        self.assertEqual(expect, actual)

    def test_error_values(self):
        data_list = [
            [
                '--property',
                'red',
            ],
            [
                '--property',
                '=',
            ],
            [
                '--property',
                '=red',
            ],
        ]
        for data in data_list:
            self.assertRaises(
                argparse.ArgumentTypeError, self.parser.parse_args, data
            )


class TestKeyValueAppendAction(utils.TestCase):
    def setUp(self):
        super(TestKeyValueAppendAction, self).setUp()

        self.parser = argparse.ArgumentParser()

        # Set up our typical usage
        self.parser.add_argument(
            '--hint',
            metavar='<key=value>',
            action=parseractions.KeyValueAppendAction,
            help='Arbitrary key/value pairs to be sent to the scheduler for '
            'custom use',
        )

    def test_good_values(self):
        print(self.parser._get_optional_actions())
        results = self.parser.parse_args(
            [
                '--hint',
                'same_host=a0cf03a5-d921-4877-bb5c-86d26cf818e1',
                '--hint',
                'same_host=8c19174f-4220-44f0-824a-cd1eeef10287',
                '--hint',
                'query=[>=,$free_ram_mb,1024]',
            ]
        )

        actual = getattr(results, 'hint', {})
        expect = {
            'same_host': [
                'a0cf03a5-d921-4877-bb5c-86d26cf818e1',
                '8c19174f-4220-44f0-824a-cd1eeef10287',
            ],
            'query': [
                '[>=,$free_ram_mb,1024]',
            ],
        }
        self.assertEqual(expect, actual)

    def test_error_values(self):
        data_list = [
            [
                '--hint',
                'red',
            ],
            [
                '--hint',
                '=',
            ],
            [
                '--hint',
                '=red',
            ],
        ]
        for data in data_list:
            self.assertRaises(
                argparse.ArgumentTypeError, self.parser.parse_args, data
            )


class TestMultiKeyValueAction(utils.TestCase):
    def setUp(self):
        super(TestMultiKeyValueAction, self).setUp()

        self.parser = argparse.ArgumentParser()

        # Set up our typical usage
        self.parser.add_argument(
            '--test',
            metavar='req1=xxx,req2=yyy',
            action=parseractions.MultiKeyValueAction,
            dest='test',
            default=None,
            required_keys=['req1', 'req2'],
            optional_keys=['opt1', 'opt2'],
            help='Test',
        )

    def test_good_values(self):
        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,req2=bbb',
                '--test',
                'req1=,req2=',
            ]
        )

        actual = getattr(results, 'test', [])
        expect = [
            {'req1': 'aaa', 'req2': 'bbb'},
            {'req1': '', 'req2': ''},
        ]
        self.assertCountEqual(expect, actual)

    def test_empty_required_optional(self):
        self.parser.add_argument(
            '--test-empty',
            metavar='req1=xxx,req2=yyy',
            action=parseractions.MultiKeyValueAction,
            dest='test_empty',
            default=None,
            required_keys=[],
            optional_keys=[],
            help='Test',
        )

        results = self.parser.parse_args(
            [
                '--test-empty',
                'req1=aaa,req2=bbb',
                '--test-empty',
                'req1=,req2=',
            ]
        )

        actual = getattr(results, 'test_empty', [])
        expect = [
            {'req1': 'aaa', 'req2': 'bbb'},
            {'req1': '', 'req2': ''},
        ]
        self.assertCountEqual(expect, actual)

    def test_error_values_with_comma(self):
        data_list = [
            [
                '--test',
                'mmm,nnn=zzz',
            ],
            [
                '--test',
                'nnn=zzz,=',
            ],
            [
                '--test',
                'nnn=zzz,=zzz',
            ],
        ]
        for data in data_list:
            self.assertRaises(
                argparse.ArgumentTypeError, self.parser.parse_args, data
            )

    def test_error_values_without_comma(self):
        self.assertRaises(
            argparse.ArgumentTypeError,
            self.parser.parse_args,
            [
                '--test',
                'mmmnnn',
            ],
        )

    def test_missing_key(self):
        self.assertRaises(
            argparse.ArgumentTypeError,
            self.parser.parse_args,
            [
                '--test',
                'req2=ddd',
            ],
        )

    def test_invalid_key(self):
        self.assertRaises(
            argparse.ArgumentTypeError,
            self.parser.parse_args,
            [
                '--test',
                'req1=aaa,req2=bbb,aaa=req1',
            ],
        )

    def test_required_keys_not_list(self):
        self.assertRaises(
            TypeError,
            self.parser.add_argument,
            '--test-required-dict',
            metavar='req1=xxx,req2=yyy',
            action=parseractions.MultiKeyValueAction,
            dest='test_required_dict',
            default=None,
            required_keys={'aaa': 'bbb'},
            optional_keys=['opt1', 'opt2'],
            help='Test',
        )

    def test_optional_keys_not_list(self):
        self.assertRaises(
            TypeError,
            self.parser.add_argument,
            '--test-optional-dict',
            metavar='req1=xxx,req2=yyy',
            action=parseractions.MultiKeyValueAction,
            dest='test_optional_dict',
            default=None,
            required_keys=['req1', 'req2'],
            optional_keys={'aaa': 'bbb'},
            help='Test',
        )


class TestMultiKeyValueCommaAction(utils.TestCase):
    def setUp(self):
        super(TestMultiKeyValueCommaAction, self).setUp()
        self.parser = argparse.ArgumentParser()

        # Typical usage
        self.parser.add_argument(
            '--test',
            metavar='req1=xxx,yyy',
            action=parseractions.MultiKeyValueCommaAction,
            dest='test',
            default=None,
            required_keys=['req1'],
            optional_keys=['opt2'],
            help='Test',
        )

    def test_mkvca_required(self):
        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,bbb',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {'req1': 'aaa,bbb'},
        ]
        self.assertCountEqual(expect, actual)

        results = self.parser.parse_args(
            [
                '--test',
                'req1=',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {'req1': ''},
        ]
        self.assertCountEqual(expect, actual)

        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,bbb',
                '--test',
                'req1=',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {'req1': 'aaa,bbb'},
            {'req1': ''},
        ]
        self.assertCountEqual(expect, actual)

    def test_mkvca_optional(self):
        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,bbb',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {'req1': 'aaa,bbb'},
        ]
        self.assertCountEqual(expect, actual)

        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,bbb',
                '--test',
                'req1=,opt2=ccc',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {'req1': 'aaa,bbb'},
            {'req1': '', 'opt2': 'ccc'},
        ]
        self.assertCountEqual(expect, actual)

        try:
            results = self.parser.parse_args(
                [
                    '--test',
                    'req1=aaa,bbb',
                    '--test',
                    'opt2=ccc',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertEqual(
                'Missing required keys req1.\nRequired keys are: req1',
                str(e),
            )

    def test_mkvca_multiples(self):
        results = self.parser.parse_args(
            [
                '--test',
                'req1=aaa,bbb,opt2=ccc',
            ]
        )
        actual = getattr(results, 'test', [])
        expect = [
            {
                'req1': 'aaa,bbb',
                'opt2': 'ccc',
            }
        ]
        self.assertCountEqual(expect, actual)

    def test_mkvca_no_required_optional(self):
        self.parser.add_argument(
            '--test-empty',
            metavar='req1=xxx,yyy',
            action=parseractions.MultiKeyValueCommaAction,
            dest='test_empty',
            default=None,
            required_keys=[],
            optional_keys=[],
            help='Test',
        )

        results = self.parser.parse_args(
            [
                '--test-empty',
                'req1=aaa,bbb',
            ]
        )
        actual = getattr(results, 'test_empty', [])
        expect = [
            {'req1': 'aaa,bbb'},
        ]
        self.assertCountEqual(expect, actual)

        results = self.parser.parse_args(
            [
                '--test-empty',
                'xyz=aaa,bbb',
            ]
        )

        actual = getattr(results, 'test_empty', [])
        expect = [
            {'xyz': 'aaa,bbb'},
        ]
        self.assertCountEqual(expect, actual)

    def test_mkvca_invalid_key(self):
        try:
            self.parser.parse_args(
                [
                    '--test',
                    'req1=aaa,bbb=',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertIn(
                'Invalid keys bbb specified.\nValid keys are:',
                str(e),
            )

        try:
            self.parser.parse_args(
                [
                    '--test',
                    'nnn=aaa',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertIn(
                'Invalid keys nnn specified.\nValid keys are:',
                str(e),
            )

    def test_mkvca_value_no_key(self):
        try:
            self.parser.parse_args(
                [
                    '--test',
                    'req1=aaa,=bbb',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertEqual(
                "A key must be specified before '=': =bbb",
                str(e),
            )
        try:
            self.parser.parse_args(
                [
                    '--test',
                    '=nnn',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertEqual(
                "A key must be specified before '=': =nnn",
                str(e),
            )

        try:
            self.parser.parse_args(
                [
                    '--test',
                    'nnn',
                ]
            )
            self.fail('ArgumentTypeError should be raised')
        except argparse.ArgumentTypeError as e:
            self.assertIn(
                'A key=value pair is required:',
                str(e),
            )

    def test_mkvca_required_keys_not_list(self):
        self.assertRaises(
            TypeError,
            self.parser.add_argument,
            '--test-required-dict',
            metavar='req1=xxx',
            action=parseractions.MultiKeyValueCommaAction,
            dest='test_required_dict',
            default=None,
            required_keys={'aaa': 'bbb'},
            optional_keys=['opt1', 'opt2'],
            help='Test',
        )

    def test_mkvca_optional_keys_not_list(self):
        self.assertRaises(
            TypeError,
            self.parser.add_argument,
            '--test-optional-dict',
            metavar='req1=xxx',
            action=parseractions.MultiKeyValueCommaAction,
            dest='test_optional_dict',
            default=None,
            required_keys=['req1', 'req2'],
            optional_keys={'aaa': 'bbb'},
            help='Test',
        )


class TestNonNegativeAction(utils.TestCase):
    def setUp(self):
        super(TestNonNegativeAction, self).setUp()

        self.parser = argparse.ArgumentParser()

        # Set up our typical usage
        self.parser.add_argument(
            '--foo',
            metavar='<foo>',
            type=int,
            action=parseractions.NonNegativeAction,
        )

    def test_negative_values(self):
        self.assertRaises(
            argparse.ArgumentTypeError,
            self.parser.parse_args,
            "--foo -1".split(),
        )

    def test_zero_values(self):
        results = self.parser.parse_args('--foo 0'.split())

        actual = getattr(results, 'foo', None)
        self.assertEqual(actual, 0)

    def test_positive_values(self):
        results = self.parser.parse_args('--foo 1'.split())

        actual = getattr(results, 'foo', None)
        self.assertEqual(actual, 1)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Base API Library"""
from keystoneauth1 import exceptions as ksa_exceptions
from keystoneauth1 import session as ksa_session
import requests

from osc_lib import exceptions
from osc_lib.i18n import _


class BaseAPI(object):
    """Base API wrapper for keystoneauth1.session.Session

    Encapsulate the translation between keystoneauth1.session.Session
    and requests.Session in a single layer:

    * Restore some requests.session.Session compatibility;
      keystoneauth1.session.Session.request() has the method and url
      arguments swapped from the rest of the requests-using world.
    * Provide basic endpoint handling when a Service Catalog is not
      available.

    """

    # Which service are we? Set in API-specific subclasses
    SERVICE_TYPE = ""

    # The common OpenStack microversion header
    HEADER_NAME = "OpenStack-API-Version"

    def __init__(
        self, session=None, service_type=None, endpoint=None, **kwargs
    ):
        """Base object that contains some common API objects and methods

        :param keystoneauth1.session.Session session:
            The session to be used for making the HTTP API calls.  If None,
            a default keystoneauth1.session.Session will be created.
        :param string service_type:
            API name, i.e. ``identity`` or ``compute``
        :param string endpoint:
            An optional URL to be used as the base for API requests on
            this API.
        :param kwargs:
            Keyword arguments passed to keystoneauth1.session.Session().
        """

        super(BaseAPI, self).__init__()

        # Create a keystoneauth1.session.Session if one is not supplied
        if not session:
            self.session = ksa_session.Session(**kwargs)
        else:
            self.session = session

        self.service_type = service_type
        self.endpoint = self._munge_endpoint(endpoint)

    def _munge_endpoint(self, endpoint):
        """Hook to allow subclasses to massage the passed-in endpoint

        Hook to massage passed-in endpoints from arbitrary sources,
        including direct user input.  By default just remove trailing
        '/' as all of our path info strings start with '/' and not all
        services can handle '//' in their URLs.

        Some subclasses will override this to do additional work, most
        likely with regard to API versions.

        :param string endpoint: The service endpoint, generally direct
                                from the service catalog.
        :return: The modified endpoint
        """

        if isinstance(endpoint, str):
            return endpoint.rstrip('/')
        else:
            return endpoint

    def _request(self, method, url, session=None, **kwargs):
        """Perform call into session

        All API calls are funneled through this method to provide a common
        place to finalize the passed URL and other things.

        :param string method:
            The HTTP method name, i.e. ``GET``, ``PUT``, etc
        :param string url:
            The API-specific portion of the URL path, or a full URL if
            ``endpoint`` was not supplied at initialization.
        :param keystoneauth1.session.Session session:
            An initialized session to override the one created at
            initialization.
        :param kwargs:
            Keyword arguments passed to requests.request().
        :return: the requests.Response object
        """

        # If session arg is supplied, use it this time, but don't save it
        if not session:
            session = self.session

        # Do the auto-endpoint magic
        if self.endpoint:
            if url:
                url = '/'.join([self.endpoint.rstrip('/'), url.lstrip('/')])
            else:
                # NOTE(dtroyer): This is left here after _munge_endpoint() is
                #                added because endpoint is public and there is
                #                no accounting for what may happen.
                url = self.endpoint.rstrip('/')
        else:
            # Pass on the lack of URL unmolested to maintain the same error
            # handling from keystoneauth: raise EndpointNotFound
            pass

        # Hack out empty headers 'cause KSA can't stomach it
        if 'headers' in kwargs and kwargs['headers'] is None:
            kwargs.pop('headers')

        # Why is ksc session backwards???
        return session.request(url, method, **kwargs)

    # The basic action methods all take a Session and return dict/lists

    def create(self, url, session=None, method=None, **params):
        """Create a new resource

        :param string url:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        :param string method:
            HTTP method (default POST)
        """

        if not method:
            method = 'POST'
        ret = self._request(method, url, session=session, **params)
        # Should this move into _requests()?
        try:
            return ret.json()
        except requests.JSONDecodeError:
            return ret

    def delete(self, url, session=None, **params):
        """Delete a resource

        :param string url:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        """

        return self._request('DELETE', url, **params)

    def list(
        self,
        path,
        session=None,
        body=None,
        detailed=False,
        headers=None,
        **params
    ):
        """Return a list of resources

        GET ${ENDPOINT}/${PATH}?${PARAMS}

        path is often the object's plural resource type

        :param string path:
            The API-specific portion of the URL path
        :param Session session:
            HTTP client session
        :param body: data that will be encoded as JSON and passed in POST
            request (GET will be sent by default)
        :param bool detailed:
            Adds '/details' to path for some APIs to return extended attributes
        :param dict headers:
            Headers dictionary to pass to requests
        :returns:
            JSON-decoded response, could be a list or a dict-wrapped-list
        """

        if detailed:
            path = '/'.join([path.rstrip('/'), 'details'])

        if body:
            ret = self._request(
                'POST',
                path,
                # service=self.service_type,
                json=body,
                params=params,
                headers=headers,
            )
        else:
            ret = self._request(
                'GET',
                path,
                # service=self.service_type,
                params=params,
                headers=headers,
            )
        try:
            return ret.json()
        except requests.JSONDecodeError:
            return ret

    # Layered actions built on top of the basic action methods do not
    # explicitly take a Session but one may still be passed in kwargs

    def find_attr(
        self,
        path,
        value=None,
        attr=None,
        resource=None,
    ):
        """Find a resource via attribute or ID

        Most APIs return a list wrapped by a dict with the resource
        name as key.  Some APIs (Identity) return a dict when a query
        string is present and there is one return value.  Take steps to
        unwrap these bodies and return a single dict without any resource
        wrappers.

        :param string path:
            The API-specific portion of the URL path
        :param string value:
            value to search for
        :param string attr:
            attribute to use for resource search
        :param string resource:
            plural of the object resource name; defaults to path

        For example:
            n = find(netclient, 'network', 'networks', 'matrix')
        """

        # Default attr is 'name'
        if attr is None:
            attr = 'name'

        # Default resource is path - in many APIs they are the same
        if resource is None:
            resource = path

        def getlist(kw):
            """Do list call, unwrap resource dict if present"""
            ret = self.list(path, **kw)
            if isinstance(ret, dict) and resource in ret:
                ret = ret[resource]
            return ret

        # Search by attribute
        kwargs = {attr: value}
        data = getlist(kwargs)
        if isinstance(data, dict):
            return data
        if len(data) == 1:
            return data[0]
        if len(data) > 1:
            msg = _("Multiple %(resource)s exist with %(attr)s='%(value)s'")
            raise exceptions.CommandError(
                msg % {'resource': resource, 'attr': attr, 'value': value}
            )

        # Search by id
        kwargs = {'id': value}
        data = getlist(kwargs)
        if len(data) == 1:
            return data[0]
        msg = _("No %(resource)s with a %(attr)s or ID of '%(value)s' found")
        raise exceptions.CommandError(
            msg % {'resource': resource, 'attr': attr, 'value': value}
        )

    def find_bulk(self, path, headers=None, **kwargs):
        """Bulk load and filter locally

        :param string path:
            The API-specific portion of the URL path
        :param kwargs:
            A dict of AVPs to match - logical AND
        :param dict headers:
            Headers dictionary to pass to requests
        :returns: list of resource dicts
        """

        items = self.list(path)
        if isinstance(items, dict):
            # strip off the enclosing dict
            key = list(items.keys())[0]
            items = items[key]

        ret = []
        for o in items:
            try:
                if all(o[attr] == kwargs[attr] for attr in kwargs.keys()):
                    ret.append(o)
            except KeyError:
                continue

        return ret

    def find_one(self, path, **kwargs):
        """Find a resource by name or ID

        :param string path:
            The API-specific portion of the URL path
        :returns:
            resource dict
        """

        bulk_list = self.find_bulk(path, **kwargs)
        num_bulk = len(bulk_list)
        if num_bulk == 0:
            msg = _("none found")
            raise exceptions.NotFound(404, msg)
        elif num_bulk > 1:
            msg = _("many found")
            raise RuntimeError(msg)
        return bulk_list[0]

    def find(
        self,
        path,
        value=None,
        attr=None,
        headers=None,
    ):
        """Find a single resource by name or ID

        :param string path:
            The API-specific portion of the URL path
        :param string value:
            search expression (required, really)
        :param string attr:
            name of attribute for secondary search
        :param dict headers:
            Headers dictionary to pass to requests
        """

        def raise_not_found():
            msg = _("%s not found") % value
            raise exceptions.NotFound(404, msg)

        try:
            ret = self._request(
                'GET',
                "/%s/%s" % (path, value),
                headers=headers,
            ).json()
            if isinstance(ret, dict):
                # strip off the enclosing dict
                key = list(ret.keys())[0]
                ret = ret[key]
        except (
            ksa_exceptions.NotFound,
            ksa_exceptions.BadRequest,
        ):
            if attr:
                kwargs = {attr: value}
                try:
                    ret = self.find_one(path, headers=headers, **kwargs)
                except (
                    exceptions.NotFound,
                    ksa_exceptions.NotFound,
                ):
                    raise_not_found()
            else:
                raise_not_found()

        return ret



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""API Utilities Library Tests"""

import copy

from osc_lib.api import api
from osc_lib.api import utils as api_utils
from osc_lib.tests.api import fakes as api_fakes


class TestBaseAPIFilter(api_fakes.TestSession):
    """The filters can be tested independently"""

    def setUp(self):
        super(TestBaseAPIFilter, self).setUp()
        self.api = api.BaseAPI(
            session=self.sess,
            endpoint=self.BASE_URL,
        )

        self.input_list = [
            api_fakes.RESP_ITEM_1,
            api_fakes.RESP_ITEM_2,
            api_fakes.RESP_ITEM_3,
        ]

    def test_simple_filter_none(self):
        output = api_utils.simple_filter()
        self.assertIsNone(output)

    def test_simple_filter_no_attr(self):
        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
        )
        self.assertEqual(self.input_list, output)

    def test_simple_filter_attr_only(self):
        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='status',
        )
        self.assertEqual(self.input_list, output)

    def test_simple_filter_attr_value(self):
        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='status',
            value='',
        )
        self.assertEqual([], output)

        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='status',
            value='UP',
        )
        self.assertEqual(
            [api_fakes.RESP_ITEM_1, api_fakes.RESP_ITEM_3],
            output,
        )

        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='fred',
            value='UP',
        )
        self.assertEqual([], output)

    def test_simple_filter_prop_attr_only(self):
        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='b',
            property_field='props',
        )
        self.assertEqual(self.input_list, output)

        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='status',
            property_field='props',
        )
        self.assertEqual(self.input_list, output)

    def test_simple_filter_prop_attr_value(self):
        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='b',
            value=2,
            property_field='props',
        )
        self.assertEqual(
            [api_fakes.RESP_ITEM_1, api_fakes.RESP_ITEM_2],
            output,
        )

        output = api_utils.simple_filter(
            copy.deepcopy(self.input_list),
            attr='b',
            value=9,
            property_field='props',
        )
        self.assertEqual([], output)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Timing Implementation"""

from osc_lib.command import command


class Timing(command.Lister):
    """Show timing data"""

    def take_action(self, parsed_args):
        column_headers = (
            'URL',
            'Seconds',
        )

        results = []
        total = 0.0
        for td in self.app.timing_data:
            sec = td.elapsed.total_seconds()
            total += sec
            results.append((td.method + ' ' + td.url, sec))
        results.append(('Total', total))
        return (
            column_headers,
            results,
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib.cli import client_config
from osc_lib.tests import utils


class TestOSCConfig(utils.TestCase):
    def setUp(self):
        super(TestOSCConfig, self).setUp()

        self.cloud = client_config.OSC_Config()

    def test_auth_select_default_plugin(self):
        config = {
            'auth_type': 'admin_token',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('admin_token', ret_config['auth_type'])

    def test_auth_select_default_plugin_password(self):
        config = {
            'username': 'fred',
            'user_id': 'fr3d',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('password', ret_config['auth_type'])
        self.assertEqual('fred', ret_config['username'])
        self.assertEqual('fr3d', ret_config['user_id'])

    def test_auth_select_default_plugin_password_v2(self):
        config = {
            'identity_api_version': '2',
            'username': 'fred',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v2password', ret_config['auth_type'])
        self.assertEqual('fred', ret_config['username'])

    def test_auth_select_default_plugin_password_v2_int(self):
        config = {
            'identity_api_version': 2,
            'username': 'fred',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v2password', ret_config['auth_type'])
        self.assertEqual('fred', ret_config['username'])

    def test_auth_select_default_plugin_password_v3(self):
        config = {
            'identity_api_version': '3',
            'username': 'fred',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v3password', ret_config['auth_type'])
        self.assertEqual('fred', ret_config['username'])

    def test_auth_select_default_plugin_password_v3_int(self):
        config = {
            'identity_api_version': 3,
            'username': 'fred',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v3password', ret_config['auth_type'])
        self.assertEqual('fred', ret_config['username'])

    def test_auth_select_default_plugin_token(self):
        config = {
            'token': 'subway',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('token', ret_config['auth_type'])
        self.assertEqual('subway', ret_config['token'])

    def test_auth_select_default_plugin_token_v2(self):
        config = {
            'identity_api_version': '2.2',
            'token': 'subway',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v2token', ret_config['auth_type'])
        self.assertEqual('subway', ret_config['token'])

    def test_auth_select_default_plugin_token_v3(self):
        config = {
            'identity_api_version': '3',
            'token': 'subway',
        }
        ret_config = self.cloud._auth_select_default_plugin(config)
        self.assertEqual('v3token', ret_config['auth_type'])
        self.assertEqual('subway', ret_config['token'])

    def test_auth_v2_arguments(self):
        config = {
            'identity_api_version': '2',
            'auth_type': 'v2password',
            'auth': {
                'username': 'fred',
            },
        }
        ret_config = self.cloud._auth_v2_arguments(config)
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertNotIn('tenant_id', ret_config['auth'])
        self.assertNotIn('tenant_name', ret_config['auth'])

        config = {
            'identity_api_version': '3',
            'auth_type': 'v3password',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
            },
        }
        ret_config = self.cloud._auth_v2_arguments(config)
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertNotIn('tenant_id', ret_config['auth'])
        self.assertNotIn('tenant_name', ret_config['auth'])

        config = {
            'identity_api_version': '2',
            'auth_type': 'v2password',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
            },
        }
        ret_config = self.cloud._auth_v2_arguments(config)
        self.assertEqual('id', ret_config['auth']['tenant_id'])
        self.assertNotIn('tenant_name', ret_config['auth'])

        config = {
            'identity_api_version': '2',
            'auth_type': 'v2password',
            'auth': {
                'username': 'fred',
                'project_name': 'name',
            },
        }
        ret_config = self.cloud._auth_v2_arguments(config)
        self.assertNotIn('tenant_id', ret_config['auth'])
        self.assertEqual('name', ret_config['auth']['tenant_name'])

    def test_auth_v2_ignore_v3(self):
        config = {
            'cloud': 'testcloud',
            'identity_api_version': '2',
            'auth_type': 'v2password',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
                'project_domain_id': 'bad',
            },
        }
        ret_config = self.cloud._auth_v2_ignore_v3(config)
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertNotIn('project_domain_id', ret_config['auth'])

    def test_auth_default_domain_not_set(self):
        config = {
            'identity_api_version': '3',
            'auth_type': 'v3oidcpassword',
            'default_domain': 'default',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
            },
        }
        ret_config = self.cloud._auth_default_domain(config)
        self.assertEqual('v3oidcpassword', ret_config['auth_type'])
        self.assertEqual('default', ret_config['default_domain'])
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertNotIn('project_domain_id', ret_config['auth'])
        self.assertNotIn('user_domain_id', ret_config['auth'])

    def test_auth_default_domain_use_default(self):
        config = {
            'identity_api_version': '3',
            'auth_type': 'v3password',
            'default_domain': 'default',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
            },
        }
        ret_config = self.cloud._auth_default_domain(config)
        self.assertEqual('v3password', ret_config['auth_type'])
        self.assertEqual('default', ret_config['default_domain'])
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertEqual('default', ret_config['auth']['project_domain_id'])
        self.assertEqual('default', ret_config['auth']['user_domain_id'])

    def test_auth_default_domain_use_given(self):
        config = {
            'identity_api_version': '3',
            'auth_type': 'v3password',
            'default_domain': 'default',
            'auth': {
                'username': 'fred',
                'project_id': 'id',
                'project_domain_id': 'proj',
                'user_domain_id': 'use',
            },
        }
        ret_config = self.cloud._auth_default_domain(config)
        self.assertEqual('v3password', ret_config['auth_type'])
        self.assertEqual('default', ret_config['default_domain'])
        self.assertEqual('fred', ret_config['auth']['username'])
        self.assertEqual('proj', ret_config['auth']['project_domain_id'])
        self.assertEqual('use', ret_config['auth']['user_domain_id'])

    def test_auth_config_hook_default(self):
        config = {}
        ret_config = self.cloud.auth_config_hook(config)
        self.assertEqual('password', ret_config['auth_type'])



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
import os
import sys
from unittest import mock

import testtools

from osc_lib import shell
from osc_lib.tests import utils


DEFAULT_AUTH_URL = "http://127.0.0.1:5000/v2.0/"
DEFAULT_PROJECT_ID = "xxxx-yyyy-zzzz"
DEFAULT_PROJECT_NAME = "project"
DEFAULT_DOMAIN_ID = "aaaa-bbbb-cccc"
DEFAULT_DOMAIN_NAME = "default"
DEFAULT_USER_DOMAIN_ID = "aaaa-bbbb-cccc"
DEFAULT_USER_DOMAIN_NAME = "domain"
DEFAULT_PROJECT_DOMAIN_ID = "aaaa-bbbb-cccc"
DEFAULT_PROJECT_DOMAIN_NAME = "domain"
DEFAULT_USERNAME = "username"
DEFAULT_PASSWORD = "password"

DEFAULT_CLOUD = "altocumulus"
DEFAULT_REGION_NAME = "ZZ9_Plural_Z_Alpha"
DEFAULT_TOKEN = "token"
DEFAULT_SERVICE_URL = "http://127.0.0.1:8771/v3.0/"
DEFAULT_AUTH_PLUGIN = "v2password"
DEFAULT_INTERFACE = "internal"

DEFAULT_COMPUTE_API_VERSION = ""
DEFAULT_IDENTITY_API_VERSION = ""
DEFAULT_IMAGE_API_VERSION = ""
DEFAULT_VOLUME_API_VERSION = ""
DEFAULT_NETWORK_API_VERSION = ""

LIB_COMPUTE_API_VERSION = ""
LIB_IDENTITY_API_VERSION = ""
LIB_IMAGE_API_VERSION = ""
LIB_VOLUME_API_VERSION = ""
LIB_NETWORK_API_VERSION = ""

CLOUD_1 = {
    'clouds': {
        'scc': {
            'auth': {
                'auth_url': DEFAULT_AUTH_URL,
                'project_name': DEFAULT_PROJECT_NAME,
                'username': 'zaphod',
            },
            'region_name': 'occ-cloud,krikkit',
            'donut': 'glazed',
            'interface': 'admin',
        }
    }
}

CLOUD_2 = {
    'clouds': {
        'megacloud': {
            'cloud': 'megadodo',
            'auth': {
                'project_name': 'heart-o-gold',
                'username': 'zaphod',
            },
            'region_name': 'occ-cloud,krikkit,occ-env',
            'log_file': '/tmp/test_log_file',
            'log_level': 'debug',
            'cert': 'mycert',
            'key': 'mickey',
        }
    }
}

PUBLIC_1 = {
    'public-clouds': {
        'megadodo': {
            'auth': {
                'auth_url': DEFAULT_AUTH_URL,
                'project_name': DEFAULT_PROJECT_NAME,
            },
            'region_name': 'occ-public',
            'donut': 'cake',
        }
    }
}


# The option table values is a tuple of (<value>, <test-opt>, <test-env>)
# where <value> is the test value to use, <test-opt> is True if this option
# should be tested as a CLI option and <test-env> is True of this option
# should be tested as an environment variable.

# Global options that should be parsed before shell.initialize_app() is called
global_options = {
    '--os-cloud': (DEFAULT_CLOUD, True, True),
    '--os-region-name': (DEFAULT_REGION_NAME, True, True),
    '--os-default-domain': (DEFAULT_DOMAIN_NAME, True, True),
    '--os-cacert': ('/dev/null', True, True),
    '--timing': (True, True, False),
    '--os-interface': (DEFAULT_INTERFACE, True, True),
}
if shell.osprofiler_profiler:
    global_options['--os-profile'] = ('SECRET_KEY', True, True)


class TestShellArgV(utils.TestShell):
    """Test the deferred help flag"""

    def setUp(self):
        super(TestShellArgV, self).setUp()

    def test_shell_argv(self):
        """Test argv decoding

        Python 2 does nothing with argv while Python 3 decodes it into
        Unicode before we ever see it.  We manually decode when running
        under Python 2 so verify that we get the right argv types.

        Use the argv supplied by the test runner so we get actual Python
        runtime behaviour; we only need to check the type of argv[0]
        which will alwyas be present.
        """

        with mock.patch(
            "osc_lib.shell.OpenStackShell.run",
            self.app,
        ):
            # Ensure type gets through unmolested through shell.main()
            argv = sys.argv
            shell.main(sys.argv)
            self.assertEqual(type(argv[0]), type(self.app.call_args[0][0][0]))

            # When shell.main() gets sys.argv itself it should be decoded
            shell.main()
            self.assertEqual(type(u'x'), type(self.app.call_args[0][0][0]))


class TestShellHelp(utils.TestShell):
    """Test the deferred help flag"""

    def setUp(self):
        super(TestShellHelp, self).setUp()
        self.useFixture(utils.EnvFixture())

    @testtools.skip("skip until bug 1444983 is resolved")
    def test_help_options(self):
        flag = "-h list server"
        kwargs = {
            "deferred_help": True,
        }
        with mock.patch(self.app_patch + ".initialize_app", self.app):
            _shell, _cmd = utils.make_shell(), flag
            utils.fake_execute(_shell, _cmd)

            self.assertEqual(
                kwargs["deferred_help"],
                _shell.options.deferred_help,
            )


class TestShellOptions(utils.TestShell):
    """Test the option handling by argparse and openstack.config.loader

    This covers getting the CLI options through the initial processing
    and validates the arguments to initialize_app() and occ_get_one()
    """

    def setUp(self):
        super(TestShellOptions, self).setUp()
        self.useFixture(utils.EnvFixture())

    def test_empty_auth(self):
        os.environ = {}
        self._assert_initialize_app_arg("", {})
        self._assert_cloud_region_arg("", {})

    def test_no_options(self):
        os.environ = {}
        self._assert_initialize_app_arg("", {})
        self._assert_cloud_region_arg("", {})

    def test_global_options(self):
        self._test_options_init_app(global_options)
        self._test_options_get_one_cloud(global_options)

    def test_global_env(self):
        self._test_env_init_app(global_options)
        self._test_env_get_one_cloud(global_options)


class TestShellCli(utils.TestShell):
    """Test handling of specific global options

    _shell.options is the parsed command line from argparse
    _shell.client_manager.* are the values actually used

    """

    def setUp(self):
        super(TestShellCli, self).setUp()
        env = {}
        self.useFixture(utils.EnvFixture(env.copy()))

    def test_shell_args_no_options(self):
        _shell = utils.make_shell()
        with mock.patch(
            "osc_lib.shell.OpenStackShell.initialize_app",
            self.app,
        ):
            utils.fake_execute(_shell, "list user")
            self.app.assert_called_with(["list", "user"])

    def test_shell_args_tls_options(self):
        """Test the TLS verify and CA cert file options"""
        _shell = utils.make_shell()

        # Default
        utils.fake_execute(_shell, "module list")
        self.assertIsNone(_shell.options.verify)
        self.assertIsNone(_shell.options.insecure)
        self.assertIsNone(_shell.options.cacert)
        self.assertTrue(_shell.client_manager.verify)
        self.assertIsNone(_shell.client_manager.cacert)

        # --verify
        utils.fake_execute(_shell, "--verify module list")
        self.assertTrue(_shell.options.verify)
        self.assertIsNone(_shell.options.insecure)
        self.assertIsNone(_shell.options.cacert)
        self.assertTrue(_shell.client_manager.verify)
        self.assertIsNone(_shell.client_manager.cacert)

        # --insecure
        utils.fake_execute(_shell, "--insecure module list")
        self.assertIsNone(_shell.options.verify)
        self.assertTrue(_shell.options.insecure)
        self.assertIsNone(_shell.options.cacert)
        self.assertFalse(_shell.client_manager.verify)
        self.assertIsNone(_shell.client_manager.cacert)

        # --os-cacert
        utils.fake_execute(_shell, "--os-cacert foo module list")
        self.assertIsNone(_shell.options.verify)
        self.assertIsNone(_shell.options.insecure)
        self.assertEqual('foo', _shell.options.cacert)
        self.assertEqual('foo', _shell.client_manager.verify)
        self.assertEqual('foo', _shell.client_manager.cacert)

        # --os-cacert and --verify
        utils.fake_execute(_shell, "--os-cacert foo --verify module list")
        self.assertTrue(_shell.options.verify)
        self.assertIsNone(_shell.options.insecure)
        self.assertEqual('foo', _shell.options.cacert)
        self.assertEqual('foo', _shell.client_manager.verify)
        self.assertEqual('foo', _shell.client_manager.cacert)

        # --os-cacert and --insecure
        # NOTE(dtroyer): Per bug https://bugs.launchpad.net/bugs/1447784
        #                in this combination --insecure now overrides any
        #                --os-cacert setting, where before --insecure
        #                was ignored if --os-cacert was set.
        utils.fake_execute(_shell, "--os-cacert foo --insecure module list")
        self.assertIsNone(_shell.options.verify)
        self.assertTrue(_shell.options.insecure)
        self.assertEqual('foo', _shell.options.cacert)
        self.assertFalse(_shell.client_manager.verify)
        self.assertIsNone(_shell.client_manager.cacert)

    def test_shell_args_cert_options(self):
        """Test client cert options"""
        _shell = utils.make_shell()

        # Default
        utils.fake_execute(_shell, "module list")
        self.assertEqual('', _shell.options.cert)
        self.assertEqual('', _shell.options.key)
        self.assertIsNone(_shell.client_manager.cert)

        # --os-cert
        utils.fake_execute(_shell, "--os-cert mycert module list")
        self.assertEqual('mycert', _shell.options.cert)
        self.assertEqual('', _shell.options.key)
        self.assertEqual('mycert', _shell.client_manager.cert)

        # --os-key
        utils.fake_execute(_shell, "--os-key mickey module list")
        self.assertEqual('', _shell.options.cert)
        self.assertEqual('mickey', _shell.options.key)
        self.assertIsNone(_shell.client_manager.cert)

        # --os-cert and --os-key
        utils.fake_execute(
            _shell, "--os-cert mycert --os-key mickey module list"
        )
        self.assertEqual('mycert', _shell.options.cert)
        self.assertEqual('mickey', _shell.options.key)
        self.assertEqual(('mycert', 'mickey'), _shell.client_manager.cert)

    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_cloud_no_vendor(self, config_mock):
        """Test cloud config options without the vendor file"""
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_1))
        _shell = utils.make_shell()

        utils.fake_execute(
            _shell,
            "--os-cloud scc module list",
        )
        self.assertEqual(
            'scc',
            _shell.cloud.name,
        )

        # These come from clouds.yaml
        #
        # To effectively assert values from clouds.yaml,
        # use values which differ from the defaults,
        # e.g. do not use public for interface because
        # osc_lib.shell.DEFAULT_INTERFACE is public.
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            DEFAULT_PROJECT_NAME,
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )
        self.assertEqual(
            'occ-cloud',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'occ-cloud',
            _shell.client_manager.region_name,
        )
        self.assertEqual(
            'glazed',
            _shell.cloud.config['donut'],
        )
        self.assertEqual(
            'admin',
            _shell.cloud.config['interface'],
        )

        self.assertIsNone(_shell.cloud.config['cert'])
        self.assertIsNone(_shell.cloud.config['key'])
        self.assertIsNone(_shell.client_manager.cert)

    @mock.patch("openstack.config.loader.OpenStackConfig._load_vendor_file")
    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_cloud_public(self, config_mock, public_mock):
        """Test cloud config options with the vendor file"""
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_2))
        public_mock.return_value = ('file.yaml', copy.deepcopy(PUBLIC_1))
        _shell = utils.make_shell()

        utils.fake_execute(
            _shell,
            "--os-cloud megacloud module list",
        )
        self.assertEqual(
            'megacloud',
            _shell.cloud.name,
        )

        # These come from clouds-public.yaml
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            'cake',
            _shell.cloud.config['donut'],
        )

        # These come from clouds.yaml
        self.assertEqual(
            'heart-o-gold',
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )
        self.assertEqual(
            'occ-cloud',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'occ-cloud',
            _shell.client_manager.region_name,
        )

        self.assertEqual('mycert', _shell.cloud.config['cert'])
        self.assertEqual('mickey', _shell.cloud.config['key'])
        self.assertEqual(('mycert', 'mickey'), _shell.client_manager.cert)

    @mock.patch("openstack.config.loader.OpenStackConfig._load_vendor_file")
    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_precedence(self, config_mock, vendor_mock):
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_2))
        vendor_mock.return_value = ('file.yaml', copy.deepcopy(PUBLIC_1))
        _shell = utils.make_shell()

        # Test command option overriding config file value
        utils.fake_execute(
            _shell,
            "--os-cloud megacloud --os-region-name krikkit module list",
        )
        self.assertEqual(
            'megacloud',
            _shell.cloud.name,
        )

        # These come from clouds-public.yaml
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            'cake',
            _shell.cloud.config['donut'],
        )

        # These come from clouds.yaml
        self.assertEqual(
            'heart-o-gold',
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )
        self.assertEqual(
            'krikkit',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'krikkit',
            _shell.client_manager.region_name,
        )


class TestShellCliPrecedence(utils.TestShell):
    """Test option precedencr order"""

    def setUp(self):
        super(TestShellCliPrecedence, self).setUp()
        env = {
            'OS_CLOUD': 'megacloud',
            'OS_REGION_NAME': 'occ-env',
        }
        self.useFixture(utils.EnvFixture(env.copy()))

    @mock.patch("openstack.config.loader.OpenStackConfig._load_vendor_file")
    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_precedence_1(self, config_mock, vendor_mock):
        """Test environment overriding occ"""
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_2))
        vendor_mock.return_value = ('file.yaml', copy.deepcopy(PUBLIC_1))
        _shell = utils.make_shell()

        # Test env var
        utils.fake_execute(
            _shell,
            "module list",
        )
        self.assertEqual(
            'megacloud',
            _shell.cloud.name,
        )

        # These come from clouds-public.yaml
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            'cake',
            _shell.cloud.config['donut'],
        )

        # These come from clouds.yaml
        self.assertEqual(
            'heart-o-gold',
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )

        # These come from the environment
        self.assertEqual(
            'occ-env',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'occ-env',
            _shell.client_manager.region_name,
        )

    @mock.patch("openstack.config.loader.OpenStackConfig._load_vendor_file")
    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_precedence_2(self, config_mock, vendor_mock):
        """Test command line overriding environment and occ"""
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_2))
        vendor_mock.return_value = ('file.yaml', copy.deepcopy(PUBLIC_1))
        _shell = utils.make_shell()

        # Test command option overriding config file value
        utils.fake_execute(
            _shell,
            "--os-region-name krikkit list user",
        )
        self.assertEqual(
            'megacloud',
            _shell.cloud.name,
        )

        # These come from clouds-public.yaml
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            'cake',
            _shell.cloud.config['donut'],
        )

        # These come from clouds.yaml
        self.assertEqual(
            'heart-o-gold',
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )

        # These come from the command line
        self.assertEqual(
            'krikkit',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'krikkit',
            _shell.client_manager.region_name,
        )

    @mock.patch("openstack.config.loader.OpenStackConfig._load_vendor_file")
    @mock.patch("openstack.config.loader.OpenStackConfig._load_config_file")
    def test_shell_args_precedence_3(self, config_mock, vendor_mock):
        """Test command line overriding environment and occ"""
        config_mock.return_value = ('file.yaml', copy.deepcopy(CLOUD_1))
        vendor_mock.return_value = ('file.yaml', copy.deepcopy(PUBLIC_1))
        _shell = utils.make_shell()

        # Test command option overriding config file value
        utils.fake_execute(
            _shell,
            "--os-cloud scc --os-region-name krikkit list user",
        )
        self.assertEqual(
            'scc',
            _shell.cloud.name,
        )

        # These come from clouds-public.yaml
        self.assertEqual(
            DEFAULT_AUTH_URL,
            _shell.cloud.config['auth']['auth_url'],
        )
        self.assertEqual(
            'glazed',
            _shell.cloud.config['donut'],
        )

        # These come from clouds.yaml
        self.assertEqual(
            DEFAULT_PROJECT_NAME,
            _shell.cloud.config['auth']['project_name'],
        )
        self.assertEqual(
            'zaphod',
            _shell.cloud.config['auth']['username'],
        )

        # These come from the command line
        self.assertEqual(
            'krikkit',
            _shell.cloud.config['region_name'],
        )
        self.assertEqual(
            'krikkit',
            _shell.client_manager.region_name,
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from osc_lib.tests import utils as test_utils
from osc_lib.utils import columns as column_utils


class TestColumnUtils(test_utils.TestCase):
    def test_get_column_definitions(self):
        attr_map = (
            ('id', 'ID', column_utils.LIST_BOTH),
            ('tenant_id', 'Project', column_utils.LIST_LONG_ONLY),
            ('name', 'Name', column_utils.LIST_BOTH),
            ('summary', 'Summary', column_utils.LIST_SHORT_ONLY),
        )
        headers, columns = column_utils.get_column_definitions(
            attr_map, long_listing=False
        )
        self.assertEqual(['id', 'name', 'summary'], columns)
        self.assertEqual(['ID', 'Name', 'Summary'], headers)

    def test_get_column_definitions_long(self):
        attr_map = (
            ('id', 'ID', column_utils.LIST_BOTH),
            ('tenant_id', 'Project', column_utils.LIST_LONG_ONLY),
            ('name', 'Name', column_utils.LIST_BOTH),
            ('summary', 'Summary', column_utils.LIST_SHORT_ONLY),
        )
        headers, columns = column_utils.get_column_definitions(
            attr_map, long_listing=True
        )
        self.assertEqual(['id', 'tenant_id', 'name'], columns)
        self.assertEqual(['ID', 'Project', 'Name'], headers)

    def test_get_columns(self):
        item = {
            'id': 'test-id',
            'tenant_id': 'test-tenant_id',
            # 'name' is not included
            'foo': 'bar',  # unknown attribute
        }
        attr_map = (
            ('id', 'ID', column_utils.LIST_BOTH),
            ('tenant_id', 'Project', column_utils.LIST_LONG_ONLY),
            ('name', 'Name', column_utils.LIST_BOTH),
        )
        columns, display_names = column_utils.get_columns(item, attr_map)
        self.assertEqual(tuple(['id', 'tenant_id', 'foo']), columns)
        self.assertEqual(tuple(['ID', 'Project', 'foo']), display_names)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import operator

LIST_BOTH = 'both'
LIST_SHORT_ONLY = 'short_only'
LIST_LONG_ONLY = 'long_only'


def get_column_definitions(attr_map, long_listing):
    """Return table headers and column names for a listing table.

    An attribute map (attr_map) is a list of table entry definitions
    and the format of the map is as follows:

    :param attr_map: a list of table entry definitions.
        Each entry should be a tuple consisting of
        (API attribute name, header name, listing mode). For example:

        .. code-block:: python

           (('id', 'ID', LIST_BOTH),
            ('name', 'Name', LIST_BOTH),
            ('tenant_id', 'Project', LIST_LONG_ONLY))

        The third field of each tuple must be one of LIST_BOTH,
        LIST_LONG_ONLY (a corresponding column is shown only in a long mode),
        or LIST_SHORT_ONLY (a corresponding column is shown only
        in a short mode).
    :param long_listing: A boolean value which indicates a long listing
        or not. In most cases, parsed_args.long is passed to this argument.
    :return: A tuple of a list of table headers and a list of column names.

    """

    if long_listing:
        headers = [
            hdr
            for col, hdr, listing_mode in attr_map
            if listing_mode in (LIST_BOTH, LIST_LONG_ONLY)
        ]
        columns = [
            col
            for col, hdr, listing_mode in attr_map
            if listing_mode in (LIST_BOTH, LIST_LONG_ONLY)
        ]
    else:
        headers = [
            hdr
            for col, hdr, listing_mode in attr_map
            if listing_mode
            if listing_mode in (LIST_BOTH, LIST_SHORT_ONLY)
        ]
        columns = [
            col
            for col, hdr, listing_mode in attr_map
            if listing_mode
            if listing_mode in (LIST_BOTH, LIST_SHORT_ONLY)
        ]

    return headers, columns


def get_columns(item, attr_map=None):
    """Return pair of resource attributes and corresponding display names.

    :param item: a dictionary which represents a resource.
        Keys of the dictionary are expected to be attributes of the resource.
        Values are not referred to by this method.

        .. code-block:: python

           {'id': 'myid', 'name': 'myname',
            'foo': 'bar', 'tenant_id': 'mytenan'}

    :param attr_map: a list of mapping from attribute to display name.
        The same format is used as for get_column_definitions attr_map.

        .. code-block:: python

           (('id', 'ID', LIST_BOTH),
            ('name', 'Name', LIST_BOTH),
            ('tenant_id', 'Project', LIST_LONG_ONLY))

    :return: A pair of tuple of attributes and tuple of display names.

        .. code-block:: python

           (('id', 'name', 'tenant_id', 'foo'),  # attributes
            ('ID', 'Name', 'Project', 'foo')     # display names

        Both tuples of attributes and display names are sorted by display names
        in the alphabetical order.
        Attributes not found in a given attr_map are kept as-is.
    """
    attr_map = attr_map or tuple([])
    _attr_map_dict = dict((col, hdr) for col, hdr, listing_mode in attr_map)

    columns = [
        (column, _attr_map_dict.get(column, column)) for column in item.keys()
    ]
    columns = sorted(columns, key=operator.itemgetter(1))
    return (tuple(col[0] for col in columns), tuple(col[1] for col in columns))



#   Copyright 2012-2013 OpenStack Foundation
#   Copyright 2015 Dean Troyer
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Command-line interface to the OpenStack APIs"""

import getpass
import logging
import sys
import traceback

from cliff import app
from cliff import command
from cliff import commandmanager
from cliff import complete
from cliff import help
from oslo_utils import importutils
from oslo_utils import strutils

from osc_lib.cli import client_config as cloud_config
from osc_lib import clientmanager
from osc_lib.command import timing
from osc_lib import exceptions as exc
from osc_lib.i18n import _
from osc_lib import logs
from osc_lib import utils
from osc_lib import version

osprofiler_profiler = importutils.try_import("osprofiler.profiler")


DEFAULT_DOMAIN = 'default'
DEFAULT_INTERFACE = 'public'


def prompt_for_password(prompt=None):
    """Prompt user for a password

    Prompt for a password if stdin is a tty.
    """

    if not prompt:
        prompt = 'Password: '
    pw = None
    # If stdin is a tty, try prompting for the password
    if hasattr(sys.stdin, 'isatty') and sys.stdin.isatty():
        # Check for Ctl-D
        try:
            pw = getpass.getpass(prompt)
        except EOFError:
            pass
    # No password because we did't have a tty or nothing was entered
    if not pw:
        raise exc.CommandError(
            _(
                "No password entered, or found via"
                " --os-password or OS_PASSWORD"
            ),
        )

    return pw


class OpenStackShell(app.App):
    CONSOLE_MESSAGE_FORMAT = '%(levelname)s: %(name)s %(message)s'

    log = logging.getLogger(__name__)
    timing_data = []

    def __init__(
        self,
        description=None,
        version=None,
        command_manager=None,
        stdin=None,
        stdout=None,
        stderr=None,
        interactive_app_factory=None,
        deferred_help=False,
    ):
        # Patch command.Command to add a default auth_required = True
        command.Command.auth_required = True

        # Some commands do not need authentication
        help.HelpCommand.auth_required = False
        complete.CompleteCommand.auth_required = False

        # Slight change to the meaning of --debug
        self.DEFAULT_DEBUG_VALUE = None
        self.DEFAULT_DEBUG_HELP = 'Set debug logging and traceback on errors.'

        # Do default for positionals
        if not command_manager:
            cm = commandmanager.CommandManager('openstack.cli')
        else:
            cm = command_manager

        super(OpenStackShell, self).__init__(
            description=__doc__.strip(),
            version=version,
            command_manager=cm,
            deferred_help=True,
        )

        # Until we have command line arguments parsed, dump any stack traces
        self.dump_stack_trace = True

        # Set in subclasses
        self.api_version = None

        self.client_manager = None
        self.command_options = None

        self.do_profile = False

    def configure_logging(self):
        """Configure logging for the app."""
        self.log_configurator = logs.LogConfigurator(self.options)
        self.dump_stack_trace = self.log_configurator.dump_trace

    def run(self, argv):
        ret_val = 1
        self.command_options = argv
        try:
            ret_val = super(OpenStackShell, self).run(argv)
            return ret_val
        except Exception as e:
            if not logging.getLogger('').handlers:
                logging.basicConfig()
            if self.dump_stack_trace:
                self.log.error(traceback.format_exc())
            else:
                self.log.error('Exception raised: ' + str(e))

            return ret_val

        finally:
            self.log.info("END return value: %s", ret_val)

    def init_profile(self):
        self.do_profile = osprofiler_profiler and self.options.profile
        if self.do_profile:
            osprofiler_profiler.init(self.options.profile)

    def close_profile(self):
        if self.do_profile:
            profiler = osprofiler_profiler.get()
            trace_id = profiler.get_base_id()
            # Short ID for OpenTracing-based driver (64-bit id)
            short_id = profiler.get_shorten_id(trace_id)

            # NOTE(dbelova): let's use warning log level to see these messages
            # printed. In fact we can define custom log level here with value
            # bigger than most big default one (CRITICAL) or something like
            # that (PROFILE = 60 for instance), but not sure we need it here.
            self.log.warning("Trace ID: %s" % trace_id)
            self.log.warning(
                "Short trace ID "
                "for OpenTracing-based drivers: %s" % short_id
            )
            self.log.warning(
                "Display trace data with command:\n"
                "osprofiler trace show --html %s " % trace_id
            )

    def run_subcommand(self, argv):
        self.init_profile()
        try:
            ret_value = super(OpenStackShell, self).run_subcommand(argv)
        finally:
            self.close_profile()
        return ret_value

    def interact(self):
        self.init_profile()
        try:
            ret_value = super(OpenStackShell, self).interact()
        finally:
            self.close_profile()
        return ret_value

    def build_option_parser(self, description, version):
        parser = super(OpenStackShell, self).build_option_parser(
            description, version
        )

        # service token auth argument
        parser.add_argument(
            '--os-cloud',
            metavar='<cloud-config-name>',
            dest='cloud',
            default=utils.env('OS_CLOUD'),
            help=_('Cloud name in clouds.yaml (Env: OS_CLOUD)'),
        )
        # Global arguments
        parser.add_argument(
            '--os-region-name',
            metavar='<auth-region-name>',
            dest='region_name',
            default=utils.env('OS_REGION_NAME'),
            help=_('Authentication region name (Env: OS_REGION_NAME)'),
        )
        parser.add_argument(
            '--os-cacert',
            metavar='<ca-bundle-file>',
            dest='cacert',
            default=utils.env('OS_CACERT', default=None),
            help=_('CA certificate bundle file (Env: OS_CACERT)'),
        )
        parser.add_argument(
            '--os-cert',
            metavar='<certificate-file>',
            dest='cert',
            default=utils.env('OS_CERT'),
            help=_('Client certificate bundle file (Env: OS_CERT)'),
        )
        parser.add_argument(
            '--os-key',
            metavar='<key-file>',
            dest='key',
            default=utils.env('OS_KEY'),
            help=_('Client certificate key file (Env: OS_KEY)'),
        )
        verify_group = parser.add_mutually_exclusive_group()
        verify_group.add_argument(
            '--verify',
            action='store_true',
            default=None,
            help=_('Verify server certificate (default)'),
        )
        verify_group.add_argument(
            '--insecure',
            action='store_true',
            default=None,
            help=_('Disable server certificate verification'),
        )
        parser.add_argument(
            '--os-default-domain',
            metavar='<auth-domain>',
            dest='default_domain',
            default=utils.env('OS_DEFAULT_DOMAIN', default=DEFAULT_DOMAIN),
            help=_(
                'Default domain ID, default=%s. ' '(Env: OS_DEFAULT_DOMAIN)'
            )
            % DEFAULT_DOMAIN,
        )
        parser.add_argument(
            '--os-interface',
            metavar='<interface>',
            dest='interface',
            choices=['admin', 'public', 'internal'],
            # do not set default value inside utils.env() else user's cloud
            # config key 'interface' will be ignored. Use OSC_Config's ctor
            # option 'override_defaults' below instead.
            default=utils.env('OS_INTERFACE'),
            help=_(
                'Select an interface type.'
                ' Valid interface types: [admin, public, internal].'
                ' default=%s, (Env: OS_INTERFACE)'
            )
            % DEFAULT_INTERFACE,
        )
        parser.add_argument(
            '--os-service-provider',
            metavar='<service_provider>',
            dest='service_provider',
            default=utils.env('OS_SERVICE_PROVIDER'),
            help=_(
                'Authenticate with and perform the command on a service'
                ' provider using Keystone-to-keystone federation. Must'
                ' also specify the remote project option.'
            ),
        )
        remote_project_group = parser.add_mutually_exclusive_group()
        remote_project_group.add_argument(
            '--os-remote-project-name',
            metavar='<remote_project_name>',
            dest='remote_project_name',
            default=utils.env('OS_REMOTE_PROJECT_NAME'),
            help=_(
                'Project name when authenticating to a service provider'
                ' if using Keystone-to-Keystone federation.'
            ),
        )
        remote_project_group.add_argument(
            '--os-remote-project-id',
            metavar='<remote_project_id>',
            dest='remote_project_id',
            default=utils.env('OS_REMOTE_PROJECT_ID'),
            help=_(
                'Project ID when authenticating to a service provider'
                ' if using Keystone-to-Keystone federation.'
            ),
        )
        remote_project_domain_group = parser.add_mutually_exclusive_group()
        remote_project_domain_group.add_argument(
            '--os-remote-project-domain-name',
            metavar='<remote_project_domain_name>',
            dest='remote_project_domain_name',
            default=utils.env('OS_REMOTE_PROJECT_DOMAIN_NAME'),
            help=_(
                'Domain name of the project when authenticating to a'
                ' service provider if using Keystone-to-Keystone'
                ' federation.'
            ),
        )
        remote_project_domain_group.add_argument(
            '--os-remote-project-domain-id',
            metavar='<remote_project_domain_id>',
            dest='remote_project_domain_id',
            default=utils.env('OS_REMOTE_PROJECT_DOMAIN_ID'),
            help=_(
                'Domain ID of the project when authenticating to a'
                ' service provider if using Keystone-to-Keystone'
                ' federation.'
            ),
        )
        parser.add_argument(
            '--timing',
            default=False,
            action='store_true',
            help=_("Print API call timing info"),
        )
        parser.add_argument(
            '--os-beta-command',
            action='store_true',
            help=_("Enable beta commands which are subject to change"),
        )

        # osprofiler HMAC key argument
        if osprofiler_profiler:
            parser.add_argument(
                '--os-profile',
                metavar='hmac-key',
                dest='profile',
                default=utils.env('OS_PROFILE'),
                help=_('HMAC key for encrypting profiling context data'),
            )

        return parser
        # return clientmanager.build_plugin_option_parser(parser)

    """
    Break up initialize_app() so that overriding it in a subclass does not
    require duplicating a lot of the method

    * super()
    * _final_defaults()
    * OpenStackConfig
    * get_one
    * _load_plugins()
    * _load_commands()
    * ClientManager

    """

    def _final_defaults(self):
        # Set the default plugin to None
        # NOTE(dtroyer): This is here to set up for setting it to a default
        #                in the calling CLI
        self._auth_type = None

        # Converge project/tenant options
        project_id = getattr(self.options, 'project_id', None)
        project_name = getattr(self.options, 'project_name', None)
        tenant_id = getattr(self.options, 'tenant_id', None)
        tenant_name = getattr(self.options, 'tenant_name', None)

        # handle some v2/v3 authentication inconsistencies by just acting like
        # both the project and tenant information are both present. This can
        # go away if we stop registering all the argparse options together.
        if project_id and not tenant_id:
            self.options.tenant_id = project_id
        if project_name and not tenant_name:
            self.options.tenant_name = project_name
        if tenant_id and not project_id:
            self.options.project_id = tenant_id
        if tenant_name and not project_name:
            self.options.project_name = tenant_name

        # Save default domain
        self.default_domain = self.options.default_domain

    def _load_plugins(self):
        """Load plugins via stevedore

        osc-lib has no opinion on what plugins should be loaded
        """
        pass

    def _load_commands(self):
        """Load commands via cliff/stevedore

        osc-lib has no opinion on what commands should be loaded
        """
        pass

    def initialize_app(self, argv):
        """Global app init bits:

        * set up API versions
        * validate authentication info
        * authenticate against Identity if requested
        """

        # Parent __init__ parses argv into self.options
        super(OpenStackShell, self).initialize_app(argv)
        self.log.info(
            "START with options: %s",
            strutils.mask_password(" ".join(self.command_options)),
        )
        self.log.debug("options: %s", strutils.mask_password(self.options))

        # Callout for stuff between superclass init and o-c-c
        self._final_defaults()

        # Do configuration file handling
        try:
            self.cloud_config = cloud_config.OSC_Config(
                pw_func=prompt_for_password,
                override_defaults={
                    'interface': DEFAULT_INTERFACE,
                    'auth_type': self._auth_type,
                },
            )
        except (IOError, OSError) as e:
            self.log.critical("Could not read clouds.yaml configuration file")
            self.print_help_if_requested()
            raise e

        # TODO(thowe): Change cliff so the default value for debug
        # can be set to None.
        if not self.options.debug:
            self.options.debug = None

        # NOTE(dtroyer): Need to do this with validate=False to defer the
        #                auth plugin handling to ClientManager.setup_auth()
        self.cloud = self.cloud_config.get_one(
            cloud=self.options.cloud,
            argparse=self.options,
            validate=False,
        )

        self.log_configurator.configure(self.cloud)
        self.dump_stack_trace = self.log_configurator.dump_trace
        self.log.debug("defaults: %s", self.cloud_config.defaults)
        self.log.debug(
            "cloud cfg: %s", strutils.mask_password(self.cloud.config)
        )

        # Callout for stuff between o-c-c and ClientManager
        # self._initialize_app_2(self.options)

        self._load_plugins()

        self._load_commands()

        # Handle deferred help and exit
        self.print_help_if_requested()

        self.client_manager = clientmanager.ClientManager(
            cli_options=self.cloud,
            api_version=self.api_version,
            pw_func=prompt_for_password,
        )

    def prepare_to_run_command(self, cmd):
        """Set up auth and API versions"""
        self.log.info(
            'command: %s -> %s.%s (auth=%s)',
            getattr(cmd, 'cmd_name', '<none>'),
            cmd.__class__.__module__,
            cmd.__class__.__name__,
            cmd.auth_required,
        )

        # NOTE(dtroyer): If auth is not required for a command, skip
        #                get_one()'s validation to avoid loading plugins
        validate = cmd.auth_required

        # NOTE(dtroyer): Save the auth required state of the _current_ command
        #                in the ClientManager
        self.client_manager._auth_required = cmd.auth_required

        # Validate auth options
        self.cloud = self.cloud_config.get_one(
            cloud=self.options.cloud,
            argparse=self.options,
            validate=validate,
            app_name=self.client_manager._app_name,
            app_version=self.client_manager._app_version,
            additional_user_agent=[('osc-lib', version.version_string)],
        )
        # Push the updated args into ClientManager
        self.client_manager._cli_options = self.cloud

        if cmd.auth_required:
            self.client_manager.setup_auth()
            if hasattr(cmd, 'required_scope') and cmd.required_scope:
                # let the command decide whether we need a scoped token
                self.client_manager.validate_scope()
            # Trigger the Identity client to initialize
            self.client_manager.session.auth.auth_ref = (
                self.client_manager.auth_ref
            )
        return

    def clean_up(self, cmd, result, err):
        self.log.debug('clean_up %s: %s', cmd.__class__.__name__, err or '')

        # Close SDK connection if available to have proper cleanup there
        if hasattr(self.client_manager, "sdk_connection"):
            self.client_manager.sdk_connection.close()

        # Close session if available
        if hasattr(self.client_manager.session, "session"):
            self.client_manager.session.session.close()

        # Process collected timing data
        if self.options.timing:
            # Get session data
            self.timing_data.extend(
                self.client_manager.session.get_timings(),
            )

            # Use the Timing pseudo-command to generate the output
            tcmd = timing.Timing(self, self.options)
            tparser = tcmd.get_parser('Timing')

            # If anything other than prettytable is specified, force csv
            format = 'table'
            # Check the formatter used in the actual command
            if (
                hasattr(cmd, 'formatter')
                and cmd.formatter != cmd._formatter_plugins['table'].obj
            ):
                format = 'csv'

            sys.stdout.write('\n')
            targs = tparser.parse_args(['-f', format])
            tcmd.run(targs)


def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]
    return OpenStackShell().run(argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
